==> sems-0.8.12.tar.gz: sems-0.8.12/AmAudio.cpp <==
/*
 * $Id: AmAudio.cpp,v 1.15.2.4 2005/08/25 07:54:04 rco Exp $
 *
 * Copyright (C) 2002-2003 Fhg Fokus
 *
 * This file is part of sems, a free SIP media server.
 *
 * sems is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * sems is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "AmAudio.h"
#include "AmPlugIn.h"
#include "amci/codecs.h"
#include "log.h"

#include <stdlib.h>
#include <string.h>
#include <assert.h>


AmAudioRtpFormat::AmAudioRtpFormat(int payload)
    : AmAudioFormat(), payload(payload)
{
    codec = getCodec();
    amci_payload_t* pl = AmPlugIn::instance()->payload(payload);
    if(pl && codec){
	sample = codec->sample_size;
	channels = pl->channels;
	rate = pl->sample_rate;
    }
    else {
	ERROR("Could not find payload <%i>\n",payload);
    }
}

AmAudioFormat::AmAudioFormat()
    : sample(-1), channels(-1), rate(-1), codec(0)
{

}

AmAudioSimpleFormat::AmAudioSimpleFormat(int codec_id)
    : AmAudioFormat(), codec_id(codec_id)
{
    codec = getCodec();
    sample = codec->sample_size;
    channels = 1;
}

AmAudioFileFormat::AmAudioFileFormat(const string& name, int subtype)
    : name(name), subtype(subtype), p_subtype(0)
{
    if(subtype >= 0){
	getSubtype();
	codec = getCodec();
	if(p_subtype && codec){
	    sample = codec->sample_size;
	    rate = p_subtype->sample_rate;
	    channels = p_subtype->channels;
	}
    }
}

AmAudioFormat::~AmAudioFormat()
{
    destroyCodec();
}

bool AmAudioFormat::operator == (const AmAudioFormat& r) const
{
    return ( codec && r.codec
	     && (r.codec->id == codec->id) 
	     && (r.sample == sample)
	     && (r.channels == channels)
	     && (r.rate == rate));
}

bool AmAudioFormat::operator != (const AmAudioFormat& r) const
{
    return !(this->operator == (r));
}

void AmAudioFormat::initCodec()
{
    if( codec && codec->init && ((h_codec = (*codec->init)()) == -1) )
	ERROR("could not initialize codec %i\n",codec->id);
}

void AmAudioFormat::destroyCodec()
{
    if( codec && codec->destroy ){
	(*codec->destroy)(h_codec);
	h_codec = 0;
    }
}

amci_subtype_t*  AmAudioFileFormat::getSubtype()
{
    if(!p_subtype && !name.empty()){

	amci_inoutfmt_t* iofmt = AmPlugIn::instance()->fileFormat(name.c_str());
	if(!iofmt){
	    ERROR("AmAudioFileFormat::getSubtype: file format '%s' does not exist\n",
		  name.c_str());
	}
	else {
	    p_subtype = AmPlugIn::instance()->subtype(iofmt,subtype);
	    if(!p_subtype)
		ERROR("AmAudioFileFormat::getSubtype: subtype %i in format '%s' does not exist\n",
		      subtype,iofmt->name);
	}
    }
    return p_subtype;
}


amci_codec_t* AmAudioFormat::getCodec()
{
    if(!codec){

	int codec_id = getCodecId();
	codec = AmPlugIn::instance()->codec(codec_id);

	initCodec();
    }
    
    return codec;
}

long AmAudioFormat::getHCodec()
{
    if(!codec)
	getCodec();
    return h_codec;
}


AmAudio::AmAudio()
    : data_size(0)
{
}

AmAudio::~AmAudio()
{
}

void AmAudio::close()
{
}

int AmAudio::convert(unsigned int user_ts, unsigned char* buffer,int nb_samples, bool is_get)
{
    assert(in.get());
    assert(out.get());
    
    if(out->rate < 0) out->rate = in->rate;
    if(out->channels < 0) out->channels = in->channels;
    
    if(in->rate < 0) in->rate = out->rate;
    if(in->channels < 0) in->channels = out->channels;


    if( *in != *out )
    {
	if(in->rate != out->rate){

	    //TODO: re-sampling
	    ERROR("sorry, we don't support audio re-sampling yet (%i -> %i)\n",in->rate,out->rate);
	    return -1;
	}

	assert(in->sample > 0);
	assert(in->channels > 0);

	int buf_size = nb_samples * in->sample * in->channels;

	// Warning: buffer overrun check does not work for frame based codecs !!!
	if( (nb_samples<<2) > AUDIO_BUFFER_SIZE ){
	    ERROR("AmAudio::get: size parameter too big: max # samples = %i\n",
		  AUDIO_BUFFER_SIZE>>2);
	    return -1;
	}

	if(is_get){
	    buf_size = streamGet(user_ts,buf_size);
	    if(buf_size == -1){
		DBG("could not read from stream\n");
		return -1;
	    }
	    if(buf_size == -2){
		DBG("end of stream\n");
		return -2;
	    }
	}
	else{
	    memcpy(samples,buffer,buf_size);
	}

	amci_codec_t* codec = in->getCodec();
	if(!codec) return -1;

  	if( (codec->id != CODEC_PCM16)
  	    && (convertInFmt2Pcm16(codec,in->getHCodec(),buf_size) < 0) )
  	    return -1;

  	if( (in->channels != out->channels)
	    && (convertChannels(buf_size) < 0) )
	    return -1;

  	codec = out->getCodec();
  	if(!codec) return -1;

  	if( (codec->id != CODEC_PCM16)
  	    && (convertPcm162OutFmt(codec,out->getHCodec(),buf_size) < 0) )
  	    return -1;

	return buf_size;
    }
    else {
	int out_size = nb_samples * out->sample * out->channels;

	if(is_get)
	    return streamGet(user_ts,out_size);
	else {
	    memcpy(samples,buffer,out_size);
	    return out_size;
	}
    }
}

// returns bytes read, else -1 if error (0 is OK)
int AmAudio::get(unsigned int user_ts, unsigned char* buffer, unsigned int nb_samples)
{
    int s = convert(user_ts, buffer, nb_samples, true);
    if(s>0)
	memcpy(buffer,(unsigned char*)samples,s);
    return s;
}

// returns bytes written, else -1 if error (0 is OK)
int AmAudio::put(unsigned int user_ts, unsigned char* buffer, unsigned int nb_samples)
{
    int s = convert(user_ts, buffer, nb_samples, false);
    if(s>0)
	return streamPut(user_ts,s);
    else
	return s;
}

void AmAudio::mono2stereo(unsigned char* out_buf,unsigned char* in_buf,int& size)
{
    unsigned short* in  = (unsigned short*)in_buf;
    unsigned short* end = (unsigned short*)(in_buf + size);
    unsigned int*   out = (unsigned int*)out_buf;

    while(in != end){
	*(out++) = ((unsigned int)*in) | (((unsigned int)*in) << 16);
	in++;
    }

    size *= 2;
}

void AmAudio::stereo2mono(unsigned char* out_buf,unsigned char* in_buf,int& size)
{
    short* in  = (short*)in_buf;
    short* end = (short*)(in_buf + size);
    short* out = (short*)out_buf;
    
    while(in != end){
	*(out++) = (*in + *(in+1)) / 2;
	in += 2;
    }

    size /= 2;
}

int AmAudio::convertInFmt2Pcm16(amci_codec_t* codec,long h_codec,int& size)
{
    if(codec->type2intern){
	int s = (*codec->type2intern)(samples.back_buffer(),samples,size,
				      in->channels,in->rate,h_codec);
	if(s<0) return -1;
	size = s;
	samples.swap();
    }
    else {
	ERROR("function ptr type2intern is NULL\n");
	return -1;
    }
    
    return 0;
}

int AmAudio::convertPcm162OutFmt(amci_codec_t* codec,long h_codec,int& size)
{
    if(codec->intern2type){
	int s = (*codec->intern2type)(samples.back_buffer(),samples,size,
				      out->channels,out->rate,h_codec);
	if(s<0) return -1;
	size = s;
	samples.swap();
    }
    else {
	ERROR("function ptr intern2type is NULL\n");
	return -1;
    }
    
    return 0;
}

int AmAudio::convertChannels(int& size)
{
    if((in->channels == 1) && (out->channels == 2))
	mono2stereo(samples.back_buffer(),samples,size);
    
    else if((in->channels == 2) && (out->channels == 1))
	stereo2mono(samples.back_buffer(),samples,size);
    
    else {
	ERROR("invalid # of channels (only mono & stereo supported)\n");
	return -1;
    }
    
    samples.swap();
    return 0;
}

DblBuffer::DblBuffer()
    : active_buf(0)
{ 
}

DblBuffer::operator unsigned char*()
{
    return samples + (active_buf ? AUDIO_BUFFER_SIZE : 0);
}

unsigned char* DblBuffer::back_buffer()
{
    return samples + (active_buf ? 0 : AUDIO_BUFFER_SIZE);
}

void DblBuffer::swap()
{
    active_buf = !active_buf;
}


// returns 0 if everything's OK
// Read mode: sets input fmt & sample size 
// Write mode: needs out fmt & sample size set
int  AmAudioFile::open(const char* filename, OpenMode mode)
{
    if(fp)
	close();
    
    AmAudioFileFormat* fmt = 
	dynamic_cast<AmAudioFileFormat*>( mode == AmAudioFile::Write ? 
					  out.get() : in.get() );

    if(!fmt){
	if(mode == AmAudioFile::Write)
	    ERROR("Output file format has not been set.\n");
	else
	    ERROR("Input file format has not been set.\n");

	return -1;
    }

    iofmt = AmPlugIn::instance()->fileFormat(fmt->getName());
    if(!iofmt){
	ERROR("output format '%s' is not valid and/or supported\n",fmt->getName().c_str());
	return -1;
    }

    open_mode = mode;
    fp = fopen(filename,mode == AmAudioFile::Read ? "rb" : "wb");
    if(!fp){
	if(mode == AmAudioFile::Read){
	    if(!filename || (strlen(filename) == 0))
		ERROR("AmAudioFile::open: filename is empty.\n");
	    else
		ERROR("AmAudioFile::open: file not found: %s\n",filename);
	}
	else
	    ERROR("could not create/overwrite file: %s\n",filename);
	return -1;
    }

    amci_file_desc_t fd;
    int ret = -1;

    fd.subtype = fmt->getSubtypeId();
    fd.sample = fmt->sample;
    fd.channels = fmt->channels;
    fd.rate = fmt->rate;
	
    if( iofmt->open && !(ret = (*iofmt->open)(fp,&fd,mode)) ) {

	fmt->setSubtypeId(fd.subtype);
	fmt->sample = fd.sample;
	fmt->channels = fd.channels;
	fmt->rate = fd.rate;

	begin = ftell(fp);
    }

    return ret;
}

AmAudioFile::AmAudioFile(const string& file_fmt, int subtype)
    : AmAudio(), fp(0), begin(0), loop(false)
{
    in.reset(new AmAudioFileFormat(file_fmt,subtype));
    out.reset(new AmAudioFileFormat(file_fmt,subtype));
}

AmAudioFile::~AmAudioFile()
{
    close();
}

void AmAudioFile::close()
{
    if(fp){

	AmAudioFileFormat* fmt = 
	    dynamic_cast<AmAudioFileFormat*>(open_mode == AmAudioFile::Write ? 
					     out.get() : in.get() );

	if(fmt){
	    amci_file_desc_t fmt_desc = { fmt->getSubtypeId(), 
					  fmt->sample, 
					  fmt->rate, 
					  fmt->channels, 
					  data_size };
	    
	    if(!iofmt){
		ERROR("file format pointer not initialized: on_close will not be called\n");
	    }
	    else if(iofmt->on_close)
		(*iofmt->on_close)(fp,&fmt_desc,open_mode);
	}


	fclose(fp);
	fp = 0;
    }
}

int AmAudioFile::streamGet(unsigned int user_ts, unsigned int size)
{
    int s = fread((void*)((unsigned char*)samples),1,size,fp);
    int ret = (!ferror(fp) ? s : -1);

    if(loop.get() && (ret <= 0) && feof(fp)){

	DBG("rewinding audio file...\n");
	fseek(fp,begin,SEEK_SET);
	s = fread((void*)((unsigned char*)samples),1,size,fp);
	ret = (!ferror(fp) ? s : -1);
    }

    if(ret > 0 && s > 0 && (unsigned int)s < size){
	DBG("0-stuffing packet: adding %i bytes (packet size=%i)\n",size-s,size);
	memset((unsigned char*)samples + s,0,size-s);
	return size;
    }

    return (feof(fp) && !loop.get() ? -2 : ret);
}

int AmAudioFile::streamPut(unsigned int user_ts, unsigned int size)
{
    int s = fwrite((void*)((unsigned char*)samples),1,size,fp);
    if(s>0)
	data_size += s;
    return (!ferror(fp) ? s : -1);
}

int AmAudioFileFormat::getCodecId()
{
    if(!name.empty()){
	getSubtype();
	if(p_subtype)
	    return p_subtype->codec_id;
    }
    
    return -1;
}

int AmAudioRtpFormat::getCodecId()
{
    if( payload >= 0 ) {

	amci_payload_t* pl = AmPlugIn::instance()->payload(payload);
	if(!pl)
	    ERROR("AmAudioRtpFormat::getCodecId: could not find payload %i\n",payload);
	else 
	    return pl->codec_id;
    }

    return -1;
}

==> sems-0.9.0.tar.gz: sems-0.9.0/AmAudio.cpp <==
/*
 * $Id: AmAudio.cpp,v 1.23.4.1 2005/08/25 09:34:20 rco Exp $
 *
 * Copyright (C) 2002-2003 Fhg Fokus
 *
 * This file is part of sems, a free SIP media server.
 *
 * sems is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * sems is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "AmAudio.h"
#include "AmPlugIn.h"
#include "amci/codecs.h"
#include "log.h"

#include <stdlib.h>
#include <string.h>
#include <assert.h>

AmAudioRtpFormat::AmAudioRtpFormat(int payload, string format_parameters)
  : AmAudioFormat(), payload(payload), amci_pl(0)
{
  sdp_format_parameters = format_parameters;
  codec = getCodec();
 
  amci_payload_t* pl = getPayloadP();
  if(pl && codec){
    sample = codec->sample_size;
    channels = pl->channels;
    rate = pl->sample_rate;
  } else {
    ERROR("Could not find payload <%i>\n",payload);
  } 
}

AmAudioFormat::AmAudioFormat()
  : sample(-1), channels(-1), rate(-1), codec(0),
    frame_length(20), frame_size(160), frame_encoded_size(320)
{

}

AmAudioSimpleFormat::AmAudioSimpleFormat(int codec_id)
    : AmAudioFormat(), codec_id(codec_id)
{
    codec = getCodec();
    sample = codec->sample_size;
}

AmAudioFileFormat::AmAudioFileFormat(const string& name, int subtype)
    : name(name), subtype(subtype), p_subtype(0)
{
    getSubtype();
    codec = getCodec();
    
    if(p_subtype && codec){
	sample = codec->sample_size;
	rate = p_subtype->sample_rate;
	channels = p_subtype->channels;
	subtype = p_subtype->type;
    }
}

AmAudioFormat::~AmAudioFormat()
{
    destroyCodec();
}

bool AmAudioFormat::operator == (const AmAudioFormat& r) const
{
    return ( codec && r.codec
	     && (r.codec->id == codec->id) 
	     && (r.sample == sample)
	     && (r.channels == channels)
	     && (r.rate == rate));
}

bool AmAudioFormat::operator != (const AmAudioFormat& r) const
{
    return !(this->operator == (r));
}

void AmAudioFormat::initCodec()
{
  amci_codec_fmt_info_t fmt_i[4];

  fmt_i[0].id=0;

  if( codec && codec->init ) {
    if ((h_codec = (*codec->init)(sdp_format_parameters.c_str(), fmt_i)) == -1) {
      ERROR("could not initialize codec %i\n",codec->id);
    } else {
      string s; 
      int i=0;
	while (fmt_i[i].id) {
	  switch (fmt_i[i].id) {
	  case AMCI_FMT_FRAME_LENGTH : {
	    frame_length=fmt_i[i].value; 
	  } break;
	  case AMCI_FMT_FRAME_SIZE: {
	    frame_size=fmt_i[i].value; 
	  } break;
	  case AMCI_FMT_ENCODED_FRAME_SIZE: {
	    frame_encoded_size=fmt_i[i].value; 
	  } break;
	  default: {
	    DBG("Unknown codec format descriptor: %d\n", fmt_i[i].id);
	  } break;
	  }
	  i++;
      }
    }  
  } 
}

void AmAudioFormat::destroyCodec()
{
    if( codec && codec->destroy ){
	(*codec->destroy)(h_codec);
	h_codec = 0;
    }
}

amci_subtype_t*  AmAudioFileFormat::getSubtype()
{
    if(!p_subtype && !name.empty()){

	amci_inoutfmt_t* iofmt = AmPlugIn::instance()->fileFormat(name.c_str());
	if(!iofmt){
	    ERROR("AmAudioFileFormat::getSubtype: file format '%s' does not exist\n",
		  name.c_str());
	    throw string("AmAudioFileFormat::getSubtype: file format '%s' does not exist\n");
	}
	else {
	    p_subtype = AmPlugIn::instance()->subtype(iofmt,subtype);
	    if(!p_subtype)
		ERROR("AmAudioFileFormat::getSubtype: subtype %i in format '%s' does not exist\n",
		      subtype,iofmt->name);
	    subtype = p_subtype->type;
	}
    }
    return p_subtype;
}


amci_codec_t* AmAudioFormat::getCodec()
{

    if(!codec){
	int codec_id = getCodecId();
	codec = AmPlugIn::instance()->codec(codec_id);

	initCodec();
    }
    
    return codec;
}

long AmAudioFormat::getHCodec()
{
    if(!codec)
	getCodec();
    return h_codec;
}


AmAudio::AmAudio()
    : data_size(0)
{
}

AmAudio::~AmAudio()
{
    //close();
}

void AmAudio::close()
{
}

int AmAudio::convert(unsigned int user_ts, unsigned char* buffer,int nb_samples, bool is_get)
{

    lockInOutFmts();

    assert(in.get());
    assert(out.get());
    
    if(out->rate < 0) out->rate = in->rate;
    if(out->channels < 0) out->channels = in->channels;
    
    if(in->rate < 0) in->rate = out->rate;
    if(in->channels < 0) in->channels = out->channels;


    if( *in != *out )
    {
	if(in->rate != out->rate){
	    //TODO: re-sampling
	    ERROR("sorry, we don't support audio re-sampling yet (%i -> %i)\n",in->rate,out->rate);
	    unlockInOutFmts();
	    return -1;
	}

	assert(in->sample > 0);
	assert(in->channels > 0);

	int buf_size = nb_samples * in->sample * in->channels;

	// Warning: buffer overrun check does not work for frame based codecs !!!
	if( (nb_samples<<2) > AUDIO_BUFFER_SIZE ){
	    ERROR("AmAudio::get: size parameter too big: max # samples = %i\n",
		  AUDIO_BUFFER_SIZE>>2);
	    unlockInOutFmts();
	    return -1;
	}

	if(is_get){
	    buf_size = streamGet(user_ts,(unsigned int) buf_size);
	    if(buf_size == -1){
		DBG("could not read from stream\n");
		unlockInOutFmts();
		return -1;
	    }
	    if(buf_size == -2){
		DBG("end of stream\n");
		unlockInOutFmts();
		return -2;
	    }
	}
	else{
	    memcpy(samples,buffer,buf_size);
	}

	amci_codec_t* codec = in->getCodec();
	if(!codec)  { unlockInOutFmts(); return -1; }

  	if( (codec->id != CODEC_PCM16)
  	    && (convertInFmt2Pcm16(codec,in->getHCodec(),buf_size) < 0) )
	  { unlockInOutFmts(); return -1; }

  	if( (in->channels != out->channels)
	    && (convertChannels(buf_size) < 0) )
	  { unlockInOutFmts(); return -1; }

  	codec = out->getCodec();
  	if(!codec) { return -1; unlockInOutFmts(); }

  	if( (codec->id != CODEC_PCM16)
  	    && (convertPcm162OutFmt(codec,out->getHCodec(),buf_size) < 0) )
	  { unlockInOutFmts(); return -1;}

	unlockInOutFmts();
	return buf_size;
    }
    else {
	int out_size = nb_samples * out->sample * out->channels;

	if(is_get) {
	    out_size = streamGet(user_ts,out_size);
	} else {
	    memcpy(samples,buffer,out_size);
	}

	unlockInOutFmts();
	return out_size;
    }
}

// returns bytes read, else -1 if error (0 is OK)
int AmAudio::get(unsigned int user_ts, unsigned char* buffer, unsigned int nb_samples)
{
    int s = convert(user_ts, buffer, nb_samples, true);
    if(s>0)
	memcpy(buffer,(unsigned char*)samples,s);
    return s;
}

// returns bytes written, else -1 if error (0 is OK)
int AmAudio::put(unsigned int user_ts, unsigned char* buffer, unsigned int nb_samples)
{
    int s = convert(user_ts, buffer, nb_samples, false);
    
    if(s>0)
	return streamPut(user_ts,(unsigned int)s);
    else
	return s;
}

void AmAudio::mono2stereo(unsigned char* out_buf,unsigned char* in_buf,int& size)
{
    unsigned short* in  = (unsigned short*)in_buf;
    unsigned short* end = (unsigned short*)(in_buf + size);
    unsigned int*   out = (unsigned int*)out_buf;

    while(in != end){
	*(out++) = ((unsigned int)*in) | (((unsigned int)*in) << 16);
	in++;
    }

    size *= 2;
}

void AmAudio::stereo2mono(unsigned char* out_buf,unsigned char* in_buf,int& size)
{
    short* in  = (short*)in_buf;
    short* end = (short*)(in_buf + size);
    short* out = (short*)out_buf;
    
    while(in != end){
	*(out++) = (*in + *(in+1)) / 2;
	in += 2;
    }

    size /= 2;
}

int AmAudio::convertInFmt2Pcm16(amci_codec_t* codec,long h_codec,int& size)
{
    if(codec->type2intern){
	int s = (*codec->type2intern)(samples.back_buffer(),samples,(unsigned int) size,
				      in->channels,in->rate,h_codec);
	if(s<0) return -1;
	size = s;
	samples.swap();
    }
    else {
	ERROR("function ptr type2intern is NULL\n");
	return -1;
    }
    
    return 0;
}

int AmAudio::convertPcm162OutFmt(amci_codec_t* codec,long h_codec,int& size)
{
    if(codec->intern2type){
	int s = (*codec->intern2type)(samples.back_buffer(),samples,(unsigned int) size,
				      out->channels,out->rate,h_codec);
	if(s<0) return -1;
	size = s;
	samples.swap();
    }
    else {
	ERROR("function ptr intern2type is NULL\n");
	return -1;
    }
    
    return 0;
}

int AmAudio::convertChannels(int& size)
{
    if((in->channels == 1) && (out->channels == 2))
	mono2stereo(samples.back_buffer(),samples,size);
    
    else if((in->channels == 2) && (out->channels == 1))
	stereo2mono(samples.back_buffer(),samples,size);
    
    else {
	ERROR("invalid # of channels (only mono & stereo supported)\n");
	return -1;
    }
    
    samples.swap();
    return 0;
}

DblBuffer::DblBuffer()
    : active_buf(0)
{ 
}

DblBuffer::operator unsigned char*()
{
    return samples + (active_buf ? AUDIO_BUFFER_SIZE : 0);
}

unsigned char* DblBuffer::back_buffer()
{
    return samples + (active_buf ? 0 : AUDIO_BUFFER_SIZE);
}

void DblBuffer::swap()
{
    active_buf = !active_buf;
}


// returns 0 if everything's OK
// Read mode: sets input fmt & sample size 
// Write mode: needs out fmt & sample size set
int  AmAudioFile::open(const char* filename, OpenMode mode)
{
    if(fp)
	close();
    
    AmAudioFileFormat* fmt = 
	dynamic_cast<AmAudioFileFormat*>( mode == AmAudioFile::Write ? 
					  out.get() : in.get() );

    if(!fmt){
	if(mode == AmAudioFile::Write)
	    ERROR("Output file format has not been set.\n");
	else
	    ERROR("Input file format has not been set.\n");

	return -1;
    }

    iofmt = AmPlugIn::instance()->fileFormat(fmt->getName());
    if(!iofmt){
	ERROR("output format '%s' is not valid and/or supported\n",fmt->getName().c_str());
	return -1;
    }

    open_mode = mode;
    fp = fopen(filename,mode == AmAudioFile::Read ? "rb" : "wb");
    if(!fp){
	if(mode == AmAudioFile::Read)
	    ERROR("file not found: %s\n",filename);
	else
	    ERROR("could not create/overwrite file: %s\n",filename);
	return -1;
    }

    amci_file_desc_t fd;
    int ret = -1;

    fd.subtype = fmt->getSubtypeId();
    fd.sample = fmt->sample;
    fd.channels = fmt->channels;
    fd.rate = fmt->rate;

    if( iofmt->open && !(ret = (*iofmt->open)(fp,&fd,mode, fmt->getHCodecNoInit())) ) {
	if (mode == AmAudioFile::Read) {
	    fmt->setSubtypeId(fd.subtype);
	    fmt->sample = fd.sample;
	    fmt->channels = fd.channels;
	    fmt->rate = fd.rate;
	}
	begin = ftell(fp);
    }

    if(open_mode == AmAudioFile::Write){

	DBG("After open:\n");
	DBG("fmt::subtype = %i\n",fmt->getSubtypeId());
	DBG("fmt::sample = %i\n",fmt->sample);
	DBG("fmt::channels = %i\n",fmt->channels);
	DBG("fmt::rate = %i\n",fmt->rate);
    }

    return ret;
}

AmAudioFile::AmAudioFile(const string& file_fmt, int subtype)
    : AmAudio(), fp(0), begin(0), loop(false)
{
    in.reset(new AmAudioFileFormat(file_fmt,subtype));
    out.reset(new AmAudioFileFormat(file_fmt,subtype));
}

AmAudioFile::~AmAudioFile()
{
    close();
}

void AmAudioFile::close()
{
    if(fp){

	AmAudioFileFormat* fmt = 
	    dynamic_cast<AmAudioFileFormat*>(open_mode == AmAudioFile::Write ? 
					     out.get() : in.get() );

	if(fmt){
	    amci_file_desc_t fmt_desc = { fmt->getSubtypeId(), 
					  fmt->sample, 
					  fmt->rate, 
					  fmt->channels, 
					  data_size };
	    
	    if(!iofmt){
		ERROR("file format pointer not initialized: on_close will not be called\n");
	    }
	    else if(iofmt->on_close)
		(*iofmt->on_close)(fp,&fmt_desc,open_mode, fmt->getHCodecNoInit());
	}

	if(open_mode == AmAudioFile::Write){

	    DBG("After close:\n");
	    DBG("fmt::subtype = %i\n",fmt->getSubtypeId());
	    DBG("fmt::sample = %i\n",fmt->sample);
	    DBG("fmt::channels = %i\n",fmt->channels);
	    DBG("fmt::rate = %i\n",fmt->rate);
	}

	fclose(fp);
	fp = 0;
    }
}

string AmAudioFile::getMimeType()
{
    if(!iofmt)
	return "";
    
    return iofmt->email_content_type;
}


int AmAudioFile::streamGet(unsigned int user_ts, unsigned int size)
{
    int s = fread((void*)((unsigned char*)samples),1,size,fp);
    int ret = (!ferror(fp) ? s : -1);

    if(loop.get() && (ret <= 0) && feof(fp)){

	DBG("rewinding audio file...\n");
	fseek(fp,begin,SEEK_SET);
	s = fread((void*)((unsigned char*)samples),1,size,fp);
	ret = (!ferror(fp) ? s : -1);
    }

    if(ret > 0 && s > 0 && (unsigned int)s < size){
	DBG("0-stuffing packet: adding %i bytes (packet size=%i)\n",size-s,size);
	memset((unsigned char*)samples + s,0,size-s);
	return size;
    }

    return (feof(fp) && !loop.get() ? -2 : ret);
}

int AmAudioFile::streamPut(unsigned int user_ts, unsigned int size)
{
    int s = fwrite((void*)((unsigned char*)samples),1,size,fp);
    if(s>0)
	data_size += s;
    return (!ferror(fp) ? s : -1);
}


int AmAudioFileFormat::getCodecId()
{
  if(!name.empty()){
      getSubtype();
	if(p_subtype)
	  return p_subtype->codec_id;
  }
    
  return -1;
}

amci_payload_t* AmAudioRtpFormat::getPayloadP()
{
    if(!amci_pl)
	amci_pl = AmPlugIn::instance()->payload(payload);

    return amci_pl;
}

int AmAudioRtpFormat::getCodecId()
{
    amci_payload_t* pl = getPayloadP();
    if(!pl){
	ERROR("AmAudioRtpFormat::getCodecId: could not find payload %i\n",payload);
	return -1;
    }
    else 
	return pl->codec_id;
}
